"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import gc
import logging
import shutil
from pathlib import Path
import torch
from safetensors.torch import load_file
from torch.distributed.fsdp.fully_sharded_data_parallel import ShardingStrategy, StateDictType
from torch.utils.data import DataLoader
from sapiens_accelerator import Accelerator, FullyShardedDataParallelPlugin
from sapiens_accelerator.commands.merge import merge_command, merge_command_parser
from sapiens_accelerator.state import AcceleratorState
from sapiens_accelerator.test_utils.training import RegressionDataset
from sapiens_accelerator.utils import merge_fsdp_weights, patch_environment, save_fsdp_model
logging.basicConfig(level=logging.INFO)
parser = merge_command_parser()
class TinyModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.linear1 = torch.nn.Linear(16, 16)
        self.activation = torch.nn.ReLU()
        self.linear2 = torch.nn.Linear(16, 16)
        self.softmax = torch.nn.Softmax()
    def forward(self, x): return self.linear2(self.activation(self.linear1(x)))
def setup():
    if AcceleratorState._shared_state != {}: AcceleratorState()._reset_state()
    plugin = FullyShardedDataParallelPlugin(sharding_strategy=ShardingStrategy.FULL_SHARD, state_dict_type=StateDictType.SHARDED_STATE_DICT)
    model = TinyModel()
    with patch_environment(fsdp_auto_wrap_policy="SIZE_BASED_WRAP"): plugin.set_auto_wrap_policy(model)
    accelerator = Accelerator(fsdp_plugin=plugin)
    model = accelerator.prepare(model)
    return model, plugin, accelerator
def mock_training(accelerator, model):
    train_set = RegressionDataset(length=128, seed=42)
    train_dl = DataLoader(train_set, batch_size=16, shuffle=False)
    optimizer = torch.optim.SGD(model.parameters(), lr=0.1)
    train_dl, model, optimizer = accelerator.prepare(train_dl, model, optimizer)
    for _ in range(3):
        for batch in train_dl:
            model.zero_grad()
            output = model(batch["x"])
            loss = torch.nn.functional.mse_loss(output, batch["y"])
            accelerator.backward(loss)
            optimizer.step()
    return model
def check_weights(operation, state_1, state_2):
    for weight_1, weight_2 in zip(state_1.values(), state_2.values()):
        if str(weight_1.device) != "cuda": weight_1 = weight_1.to("cuda")
        if str(weight_2.device) != "cuda": weight_2 = weight_2.to("cuda")
        if operation == "same": assert torch.allclose(weight_1, weight_2)
        else: assert not torch.allclose(weight_1, weight_2)
def check_safetensors_weights(path, model):
    safe_state_dict = load_file(path / "model.safetensors")
    safe_loaded_model = TinyModel()
    check_weights("diff", model.state_dict(), safe_loaded_model.state_dict())
    safe_loaded_model.load_state_dict(safe_state_dict)
    check_weights("same", model.state_dict(), safe_loaded_model.state_dict())
def check_pytorch_weights(path, model):
    nonsafe_state_dict = torch.load(path / "pytorch_model.bin")
    nonsafe_loaded_model = TinyModel()
    check_weights("diff", model.state_dict(), nonsafe_loaded_model.state_dict())
    nonsafe_loaded_model.load_state_dict(nonsafe_state_dict)
    check_weights("same", model.state_dict(), nonsafe_loaded_model.state_dict())
def test_merge_weights_safetensors(model, path):
    merge_fsdp_weights(path / "pytorch_model_fsdp_0", path, safe_serialization=True)
    check_safetensors_weights(path, model)
def test_merge_weights_command_safetensors(model, path):
    args = parser.parse_args([str(path / "pytorch_model_fsdp_0"), str(path)])
    merge_command(args)
    check_safetensors_weights(path, model)
def test_merge_weights_pytorch(model, path):
    merge_fsdp_weights(path / "pytorch_model_fsdp_0", path, safe_serialization=False)
    check_pytorch_weights(path, model)
def test_merge_weights_command_pytorch(model, path):
    args = parser.parse_args([str(path / "pytorch_model_fsdp_0"), str(path), "--unsafe_serialization"])
    merge_command(args)
    check_pytorch_weights(path, model)
if __name__ == "__main__":
    model, plugin, accelerator = setup()
    if accelerator.num_processes > 1:
        try:
            out_path = Path("test_merge_weights_fsdp_weights")
            if not out_path.exists(): out_path.mkdir(parents=True, exist_ok=True)
            model = mock_training(accelerator, model)
            accelerator.wait_for_everyone()
            gc.collect()
            save_fsdp_model(plugin, accelerator, model, out_path)
            accelerator.wait_for_everyone()
            test_merge_weights_safetensors(model, out_path)
            test_merge_weights_command_safetensors(model, out_path)
            test_merge_weights_pytorch(model, out_path)
            test_merge_weights_command_pytorch(model, out_path)
        except Exception: raise
        finally:
            if accelerator.is_main_process: shutil.rmtree(out_path)
            accelerator.wait_for_everyone()
            accelerator.end_training()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import os
import shutil
from pathlib import Path
import torch
from ..logging import get_logger
from .constants import FSDP_MODEL_NAME, OPTIMIZER_NAME, SAFE_WEIGHTS_NAME, WEIGHTS_NAME
from .modeling import is_peft_model
from .other import save
from .versions import is_torch_version
logger = get_logger(__name__)
def enable_fsdp_ram_efficient_loading():
    if "SAPIENS_ACCELERATOR_USE_FSDP" not in os.environ: os.environ["SAPIENS_ACCELERATOR_USE_FSDP"] = "True"
    os.environ["FSDP_CPU_RAM_EFFICIENT_LOADING"] = "True"
def disable_fsdp_ram_efficient_loading(): os.environ["FSDP_CPU_RAM_EFFICIENT_LOADING"] = "False"
def _get_model_state_dict(model, adapter_only=False):
    if adapter_only and is_peft_model(model):
        from peft import get_peft_model_state_dict
        return get_peft_model_state_dict(model, adapter_name=model.active_adapter)
    else: return model.state_dict()
def _set_model_state_dict(model, state_dict, adapter_only=False):
    if adapter_only and is_peft_model(model):
        from peft import set_peft_model_state_dict
        return set_peft_model_state_dict(model, state_dict, adapter_name=model.active_adapter)
    else: return model.load_state_dict(state_dict)
def save_fsdp_model(fsdp_plugin, accelerator, model, output_dir, model_index=0, adapter_only=False):
    import torch.distributed.checkpoint as dist_cp
    from torch.distributed.checkpoint.default_planner import DefaultSavePlanner
    from torch.distributed.fsdp.fully_sharded_data_parallel import FullyShardedDataParallel as FSDP
    from torch.distributed.fsdp.fully_sharded_data_parallel import StateDictType
    os.makedirs(output_dir, exist_ok=True)
    if fsdp_plugin.state_dict_type == StateDictType.FULL_STATE_DICT:
        is_multi_process = accelerator.num_processes > 1
        fsdp_plugin.state_dict_config.offload_to_cpu = is_multi_process
        fsdp_plugin.state_dict_config.rank0_only = is_multi_process
    with FSDP.state_dict_type(model, fsdp_plugin.state_dict_type, fsdp_plugin.state_dict_config, fsdp_plugin.optim_state_dict_config):
        state_dict = _get_model_state_dict(model, adapter_only=adapter_only)
        if fsdp_plugin.state_dict_type == StateDictType.FULL_STATE_DICT:
            weights_name = f"{FSDP_MODEL_NAME}.bin" if model_index == 0 else f"{FSDP_MODEL_NAME}_{model_index}.bin"
            output_model_file = os.path.join(output_dir, weights_name)
            if accelerator.process_index == 0: torch.save(state_dict, output_model_file)
        elif fsdp_plugin.state_dict_type == StateDictType.LOCAL_STATE_DICT:
            weights_name = (f"{FSDP_MODEL_NAME}_rank{accelerator.process_index}.bin" if model_index == 0 else f"{FSDP_MODEL_NAME}_{model_index}_rank{accelerator.process_index}.bin")
            output_model_file = os.path.join(output_dir, weights_name)
            torch.save(state_dict, output_model_file)
        elif fsdp_plugin.state_dict_type == StateDictType.SHARDED_STATE_DICT:
            ckpt_dir = os.path.join(output_dir, f"{FSDP_MODEL_NAME}_{model_index}")
            os.makedirs(ckpt_dir, exist_ok=True)
            state_dict = {"model": state_dict}
            dist_cp.save_state_dict(state_dict=state_dict, storage_writer=dist_cp.FileSystemWriter(ckpt_dir), planner=DefaultSavePlanner())
def load_fsdp_model(fsdp_plugin, accelerator, model, input_dir, model_index=0, adapter_only=False):
    import torch.distributed.checkpoint as dist_cp
    from torch.distributed.checkpoint.default_planner import DefaultLoadPlanner
    from torch.distributed.fsdp.fully_sharded_data_parallel import FullyShardedDataParallel as FSDP
    from torch.distributed.fsdp.fully_sharded_data_parallel import StateDictType
    accelerator.wait_for_everyone()
    if fsdp_plugin.state_dict_type == StateDictType.FULL_STATE_DICT:
        is_multi_process = accelerator.num_processes > 1
        fsdp_plugin.state_dict_config.offload_to_cpu = is_multi_process
        fsdp_plugin.state_dict_config.rank0_only = is_multi_process
    with FSDP.state_dict_type(model, fsdp_plugin.state_dict_type, fsdp_plugin.state_dict_config, fsdp_plugin.optim_state_dict_config):
        if fsdp_plugin.state_dict_type == StateDictType.FULL_STATE_DICT:
            if type(model) is not FSDP and accelerator.process_index != 0:
                if not fsdp_plugin.sync_module_states: raise ValueError("Set the `sync_module_states` flag to `True` so that model states are synced across processes when initializing FSDP object")
                return
            weights_name = f"{FSDP_MODEL_NAME}.bin" if model_index == 0 else f"{FSDP_MODEL_NAME}_{model_index}.bin"
            input_model_file = os.path.join(input_dir, weights_name)
            state_dict = torch.load(input_model_file)
        elif fsdp_plugin.state_dict_type == StateDictType.LOCAL_STATE_DICT:
            weights_name = (f"{FSDP_MODEL_NAME}_rank{accelerator.process_index}.bin" if model_index == 0 else f"{FSDP_MODEL_NAME}_{model_index}_rank{accelerator.process_index}.bin")
            input_model_file = os.path.join(input_dir, weights_name)
            state_dict = torch.load(input_model_file)
        elif fsdp_plugin.state_dict_type == StateDictType.SHARDED_STATE_DICT:
            ckpt_dir = (os.path.join(input_dir, f"{FSDP_MODEL_NAME}_{model_index}") if f"{FSDP_MODEL_NAME}" not in input_dir else input_dir)
            state_dict = {"model": _get_model_state_dict(model, adapter_only=adapter_only)}
            dist_cp.load_state_dict(state_dict=state_dict, storage_reader=dist_cp.FileSystemReader(ckpt_dir), planner=DefaultLoadPlanner())
            state_dict = state_dict["model"]
        load_result = _set_model_state_dict(model, state_dict, adapter_only=adapter_only)
    return load_result
def save_fsdp_optimizer(fsdp_plugin, accelerator, optimizer, model, output_dir, optimizer_index=0):
    import torch.distributed.checkpoint as dist_cp
    from torch.distributed.checkpoint.default_planner import DefaultSavePlanner
    from torch.distributed.fsdp.fully_sharded_data_parallel import FullyShardedDataParallel as FSDP
    from torch.distributed.fsdp.fully_sharded_data_parallel import StateDictType
    os.makedirs(output_dir, exist_ok=True)
    with FSDP.state_dict_type(model, fsdp_plugin.state_dict_type, fsdp_plugin.state_dict_config, fsdp_plugin.optim_state_dict_config):
        optim_state = FSDP.optim_state_dict(model, optimizer)
        if fsdp_plugin.state_dict_type == StateDictType.FULL_STATE_DICT:
            if accelerator.process_index == 0:
                optim_state_name = (f"{OPTIMIZER_NAME}.bin" if optimizer_index == 0 else f"{OPTIMIZER_NAME}_{optimizer_index}.bin")
                output_optimizer_file = os.path.join(output_dir, optim_state_name)
                torch.save(optim_state, output_optimizer_file)
        else:
            ckpt_dir = os.path.join(output_dir, f"{OPTIMIZER_NAME}_{optimizer_index}")
            os.makedirs(ckpt_dir, exist_ok=True)
            dist_cp.save_state_dict(state_dict={"optimizer": optim_state}, storage_writer=dist_cp.FileSystemWriter(ckpt_dir), planner=DefaultSavePlanner())
def load_fsdp_optimizer(fsdp_plugin, accelerator, optimizer, model, input_dir, optimizer_index=0, adapter_only=False):
    import torch.distributed.checkpoint as dist_cp
    from torch.distributed.checkpoint.optimizer import load_sharded_optimizer_state_dict
    from torch.distributed.fsdp.fully_sharded_data_parallel import FullyShardedDataParallel as FSDP
    from torch.distributed.fsdp.fully_sharded_data_parallel import StateDictType
    accelerator.wait_for_everyone()
    with FSDP.state_dict_type(model, fsdp_plugin.state_dict_type, fsdp_plugin.state_dict_config, fsdp_plugin.optim_state_dict_config):
        if fsdp_plugin.state_dict_type == StateDictType.FULL_STATE_DICT:
            optim_state = None
            if accelerator.process_index == 0 or not fsdp_plugin.optim_state_dict_config.rank0_only:
                optimizer_name = (f"{OPTIMIZER_NAME}.bin" if optimizer_index == 0 else f"{OPTIMIZER_NAME}_{optimizer_index}.bin")
                input_optimizer_file = os.path.join(input_dir, optimizer_name)
                optim_state = torch.load(input_optimizer_file)
        else:
            ckpt_dir = (os.path.join(input_dir, f"{OPTIMIZER_NAME}_{optimizer_index}") if f"{OPTIMIZER_NAME}" not in input_dir else input_dir)
            optim_state = load_sharded_optimizer_state_dict(model_state_dict=_get_model_state_dict(model, adapter_only=adapter_only), optimizer_key="optimizer", storage_reader=dist_cp.FileSystemReader(ckpt_dir))
            optim_state = optim_state["optimizer"]
        flattened_osd = FSDP.optim_state_dict_to_load(model=model, optim=optimizer, optim_state_dict=optim_state)
        optimizer.load_state_dict(flattened_osd)
def _distributed_checkpoint_to_merged_weights(checkpoint_dir: str, save_path: str, safe_serialization: bool = True):
    import torch.distributed.checkpoint as dist_cp
    import torch.distributed.checkpoint.format_utils as dist_cp_format_utils
    state_dict = {}
    save_path = Path(save_path)
    save_path.mkdir(exist_ok=True)
    dist_cp_format_utils._load_state_dict(state_dict, storage_reader=dist_cp.FileSystemReader(checkpoint_dir), planner=dist_cp_format_utils._EmptyStateDictLoadPlanner(), no_dist=True)
    save_path = save_path / SAFE_WEIGHTS_NAME if safe_serialization else save_path / WEIGHTS_NAME
    if len(state_dict.keys()) == 1: state_dict = state_dict[list(state_dict)[0]]
    save(state_dict, save_path, safe_serialization=safe_serialization)
    return save_path
def merge_fsdp_weights(checkpoint_dir: str, output_path: str, safe_serialization: bool = True, remove_checkpoint_dir: bool = False):
    checkpoint_dir = Path(checkpoint_dir)
    from sapiens_accelerator.state import PartialState
    if not is_torch_version(">=", "2.3.0"): raise ValueError("`merge_fsdp_weights` requires PyTorch >= 2.3.0`")
    if not checkpoint_dir.exists():
        model_path_exists = (checkpoint_dir / "pytorch_model_fsdp_0").exists()
        optimizer_path_exists = (checkpoint_dir / "optimizer_0").exists()
        err = f"Tried to load from {checkpoint_dir} but couldn't find a valid metadata file."
        if model_path_exists and optimizer_path_exists:
            err += " However, potential model and optimizer checkpoint directories exist."
            err += f"Please pass in either {checkpoint_dir}/pytorch_model_fsdp_0 or {checkpoint_dir}/optimizer_0"
            err += "instead."
        elif model_path_exists:
            err += " However, a potential model checkpoint directory exists."
            err += f"Please try passing in {checkpoint_dir}/pytorch_model_fsdp_0 instead."
        elif optimizer_path_exists:
            err += " However, a potential optimizer checkpoint directory exists."
            err += f"Please try passing in {checkpoint_dir}/optimizer_0 instead."
        raise ValueError(err)
    state = PartialState()
    if state.is_main_process:
        save_path = _distributed_checkpoint_to_merged_weights(checkpoint_dir, output_path, safe_serialization)
        if remove_checkpoint_dir: shutil.rmtree(checkpoint_dir)
    state.wait_for_everyone()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import warnings
from .state import AcceleratorState, GradientState
class SapiensAcceleratordScheduler:
    def __init__(self, scheduler, optimizers, step_with_optimizer: bool = True, split_batches: bool = False):
        self.scheduler = scheduler
        self.optimizers = optimizers if isinstance(optimizers, (list, tuple)) else [optimizers]
        self.split_batches = split_batches
        self.step_with_optimizer = step_with_optimizer
        self.gradient_state = GradientState()
    def step(self, *args, **kwargs):
        if not self.step_with_optimizer:
            self.scheduler.step(*args, **kwargs)
            return
        if not self.gradient_state.sync_gradients:
            if self.gradient_state.adjust_scheduler: self.scheduler._step_count += 1
            return
        for opt in self.optimizers:
            if opt.step_was_skipped: return
        if self.split_batches: self.scheduler.step(*args, **kwargs)
        else:
            num_processes = AcceleratorState().num_processes
            for _ in range(num_processes):
                if hasattr(self.scheduler, "total_steps"):
                    if self.scheduler._step_count <= self.scheduler.total_steps: self.scheduler.step(*args, **kwargs)
                else: self.scheduler.step(*args, **kwargs)
    def get_last_lr(self): return self.scheduler.get_last_lr()
    def state_dict(self): return self.scheduler.state_dict()
    def load_state_dict(self, state_dict): self.scheduler.load_state_dict(state_dict)
    def get_lr(self): return self.scheduler.get_lr()
    def print_lr(self, *args, **kwargs): return self.scheduler.print_lr(*args, **kwargs)
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
from .constants import (MITA_PROFILING_AVAILABLE_PYTORCH_VERSION, MODEL_NAME, OPTIMIZER_NAME, PROFILE_PATTERN_NAME, RNG_STATE_NAME, SAFE_MODEL_NAME,
SAFE_WEIGHTS_INDEX_NAME, SAFE_WEIGHTS_NAME, SAFE_WEIGHTS_PATTERN_NAME, SAMPLER_NAME, SCALER_NAME, SCHEDULER_NAME, TORCH_DISTRIBUTED_OPERATION_TYPES,
TORCH_LAUNCH_PARAMS, WEIGHTS_INDEX_NAME, WEIGHTS_NAME, WEIGHTS_PATTERN_NAME, XPU_PROFILING_AVAILABLE_PYTORCH_VERSION)
from .dataclasses import (AutocastKwargs, BnbQuantizationConfig, ComputeEnvironment, CustomDtype, DataLoaderConfiguration, DDPCommunicationHookType,
DeepSpeedPlugin, DistributedDataParallelKwargs, DistributedType, DynamoBackend, FP8RecipeKwargs, FullyShardedDataParallelPlugin, GradientAccumulationPlugin,
GradScalerKwargs, InitProcessGroupKwargs, KwargsHandler, LoggerType, MegatronLMPlugin, PrecisionType, ProfileKwargs, ProjectConfiguration, RNGType,
SageMakerDistributedType, TensorInformation, TorchDynamoPlugin, add_model_config_to_megatron_parser)
from .environment import (are_libraries_initialized, check_cuda_p2p_ib_support, check_fp8_capability, convert_dict_to_env_variables, get_cpu_distributed_information,
get_gpu_info, get_int_from_env, parse_choice_from_env, parse_flag_from_env, set_numa_affinity, str_to_bool)
from .imports import (deepspeed_required, get_ccl_version, is_4bit_sapiens_available, is_8bit_sapiens_available, is_aim_available, is_bf16_available, is_sapiens_machine_multi_backend_available,
is_sapiens_available, is_boto3_available, is_ccl_available, is_clearml_available, is_comet_ml_available, is_cuda_available, is_datasets_available, is_deepspeed_available,
is_dvclive_available, is_fp8_available, is_import_timer_available, is_ipex_available, is_lomo_available, is_megatron_lm_available, is_mlflow_available,
is_mlu_available, is_mps_available, is_msamp_available, is_musa_available, is_npu_available, is_pandas_available, is_peft_available, is_pippy_available,
is_pynvml_available, is_pytest_available, is_rich_available, is_sagemaker_available, is_schedulefree_available, is_tensorboard_available, is_timm_available,
is_torch_xla_available, is_torchdata_available, is_torchdata_stateful_dataloader_available, is_torchvision_available, is_transformer_engine_available,
is_transformers_available, is_triton_available, is_wandb_available, is_xpu_available)
from .modeling import (calculate_maximum_sizes, check_device_map, check_tied_parameters_in_config, check_tied_parameters_on_same_device, compute_module_sizes,
convert_file_size_to_int, dtype_byte_size, find_tied_parameters, get_balanced_memory, get_max_layer_size, get_max_memory, get_mixed_precision_context_manager,
id_tensor_storage, infer_auto_device_map, is_peft_model, load_checkpoint_in_model, load_offloaded_weights, load_state_dict, named_module_tensors, retie_parameters, set_module_tensor_to_device)
from .offload import (OffloadedWeightsLoader, PrefixedDataset, extract_submodules_state_dict, load_offloaded_weight, offload_state_dict, offload_weight, save_offload_index)
from .operations import (CannotPadNestedTensorWarning, GatheredParameters, broadcast, broadcast_object_list, concatenate, convert_outputs_to_fp32,
convert_to_fp32, copy_tensor_to_devices, find_batch_size, find_device, gather, gather_object, get_data_structure, honor_type, ignorant_find_batch_size,
initialize_tensors, is_namedtuple, is_tensor_information, is_torch_tensor, listify, pad_across_processes, pad_input_tensors, recursively_apply,
reduce, send_to_device, slice_tensors)
from .versions import compare_versions, is_torch_version
if is_deepspeed_available(): from .deepspeed import (DeepSpeedEngineWrapper, DeepSpeedOptimizerWrapper, DeepSpeedSchedulerWrapper, DummyOptim, DummyScheduler, HfDeepSpeedConfig, get_active_deepspeed_plugin)
from .sapiens import has_4bit_sapiens_layers, load_and_quantize_model
from .fsdp_utils import (disable_fsdp_ram_efficient_loading, enable_fsdp_ram_efficient_loading, load_fsdp_model, load_fsdp_optimizer, merge_fsdp_weights, save_fsdp_model, save_fsdp_optimizer)
from .launch import (PrepareForLaunch, _filter_args, prepare_deepspeed_cmd_env, prepare_multi_gpu_env, prepare_sagemager_args_inputs, prepare_simple_launcher_cmd_env, prepare_tpu)
from .megatron_lm import (AbstractTrainStep, BertTrainStep, GPTTrainStep, MegatronLMDummyDataLoader, MegatronLMDummyScheduler, T5TrainStep, avg_losses_across_data_parallel_group)
if is_megatron_lm_available():
    from .megatron_lm import (MegatronEngine, MegatronLMOptimizerWrapper, MegatronLMSchedulerWrapper, gather_across_data_parallel_groups)
    from .megatron_lm import initialize as megatron_lm_initialize
    from .megatron_lm import prepare_data_loader as megatron_lm_prepare_data_loader
    from .megatron_lm import prepare_model_optimizer_scheduler as megatron_lm_prepare_model_optimizer_scheduler
    from .megatron_lm import prepare_optimizer as megatron_lm_prepare_optimizer
    from .megatron_lm import prepare_scheduler as megatron_lm_prepare_scheduler
from .memory import find_executable_batch_size, release_memory
from .other import (check_os_kernel, clean_state_dict_for_safetensors, clear_environment, convert_bytes, extract_model_from_parallel, get_pretty_name,
is_port_in_use, merge_dicts, patch_environment, recursive_getattr, save, wait_for_everyone, write_basic_config)
from .random import set_seed, synchronize_rng_state, synchronize_rng_states
from .torch_xla import install_xla
from .tqdm import tqdm
from .transformer_engine import (apply_fp8_autowrap, contextual_fp8_autocast, convert_model, has_transformer_engine_layers)
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import contextlib
import gc
import inspect
import json
import logging
import os
import re
import shutil
import tempfile
import warnings
from collections import OrderedDict, defaultdict
from typing import Dict, List, Optional, Set, Tuple, Union
import torch
import torch.nn as nn
from ..state import AcceleratorState
from .constants import SAFE_WEIGHTS_NAME, WEIGHTS_NAME
from .dataclasses import AutocastKwargs, CustomDtype, DistributedType
from .imports import (is_mlu_available, is_mps_available, is_musa_available, is_npu_available, is_peft_available, is_torch_xla_available, is_xpu_available)
from .memory import clear_device_cache, get_xpu_available_memory
from .offload import load_offloaded_weight, offload_weight, save_offload_index
from .tqdm import is_tqdm_available, tqdm
from .versions import compare_versions, is_torch_version
if is_npu_available(check_device=False): import torch_npu
if is_mlu_available(check_device=False): import torch_mlu
if is_musa_available(check_device=False): import torch_musa
from safetensors import safe_open
from safetensors.torch import load_file as safe_load_file
WEIGHTS_INDEX_NAME = "pytorch_model.bin.index.json"
logger = logging.getLogger(__name__)
def is_peft_model(model):
    from .other import extract_model_from_parallel
    if is_peft_available(): from peft import PeftModel
    return is_peft_available() and isinstance(extract_model_from_parallel(model), PeftModel)
def check_device_same(first_device, second_device):
    if first_device.type != second_device.type: return False
    if first_device.type == "cuda" and first_device.index is None: first_device = torch.device("cuda", index=0)
    if second_device.type == "cuda" and second_device.index is None: second_device = torch.device("cuda", index=0)
    return first_device == second_device
def convert_file_size_to_int(size: Union[int, str]):
    mem_size = -1
    err_msg = (f"`size` {size} is not in a valid format. Use an integer for bytes, or a string with an unit (like '5.0GB').")
    try:
        if isinstance(size, int): mem_size = size
        elif size.upper().endswith("GIB"): mem_size = int(float(size[:-3]) * (2**30))
        elif size.upper().endswith("MIB"): mem_size = int(float(size[:-3]) * (2**20))
        elif size.upper().endswith("KIB"): mem_size = int(float(size[:-3]) * (2**10))
        elif size.upper().endswith("GB"):
            int_size = int(float(size[:-2]) * (10**9))
            mem_size = int_size // 8 if size.endswith("b") else int_size
        elif size.upper().endswith("MB"):
            int_size = int(float(size[:-2]) * (10**6))
            mem_size = int_size // 8 if size.endswith("b") else int_size
        elif size.upper().endswith("KB"):
            int_size = int(float(size[:-2]) * (10**3))
            mem_size = int_size // 8 if size.endswith("b") else int_size
    except ValueError: raise ValueError(err_msg)
    if mem_size < 0: raise ValueError(err_msg)
    return mem_size
def dtype_byte_size(dtype: torch.dtype):
    if dtype == torch.bool: return 1 / 8
    elif dtype == CustomDtype.INT2: return 1 / 4
    elif dtype == CustomDtype.INT4: return 1 / 2
    elif dtype == CustomDtype.FP8: return 1
    elif is_torch_version(">=", "2.1.0") and dtype == torch.float8_e4m3fn: return 1
    bit_search = re.search(r"[^\d](\d+)$", str(dtype))
    if bit_search is None: raise ValueError(f"`dtype` is not a valid dtype: {dtype}.")
    bit_size = int(bit_search.groups()[0])
    return bit_size // 8
def id_tensor_storage(tensor: torch.Tensor) -> Tuple[torch.device, int, int]:
    _SIZE = {torch.int64: 8, torch.float32: 4, torch.int32: 4, torch.bfloat16: 2, torch.float16: 2, torch.int16: 2, torch.uint8: 1, torch.int8: 1, torch.bool: 1, torch.float64: 8}
    try:
        storage_ptr = tensor.untyped_storage().data_ptr()
        storage_size = tensor.untyped_storage().nbytes()
    except Exception:
        try:
            storage_ptr = tensor.storage().data_ptr()
            storage_size = tensor.storage().size() * _SIZE[tensor.dtype]
        except NotImplementedError:
            storage_ptr = 0
            storage_size = tensor.nelement() * _SIZE[tensor.dtype]
    return tensor.device, storage_ptr, storage_size
def set_module_tensor_to_device(module: nn.Module, tensor_name: str, device: Union[int, str, torch.device], value: Optional[torch.Tensor] = None,
dtype: Optional[Union[str, torch.dtype]] = None, fp16_statistics: Optional[torch.HalfTensor] = None, tied_params_map: Optional[Dict[int, Dict[torch.device, torch.Tensor]]] = None):
    if "." in tensor_name:
        splits = tensor_name.split(".")
        for split in splits[:-1]:
            new_module = getattr(module, split)
            if new_module is None: raise ValueError(f"{module} has no attribute {split}.")
            module = new_module
        tensor_name = splits[-1]
    if tensor_name not in module._parameters and tensor_name not in module._buffers: raise ValueError(f"{module} does not have a parameter or a buffer named {tensor_name}.")
    is_buffer = tensor_name in module._buffers
    old_value = getattr(module, tensor_name)
    if (value is not None and tied_params_map is not None and value.data_ptr() in tied_params_map and device in tied_params_map[value.data_ptr()]):
        module._parameters[tensor_name] = tied_params_map[value.data_ptr()][device]
        return
    elif (tied_params_map is not None and old_value.data_ptr() in tied_params_map and device in tied_params_map[old_value.data_ptr()]):
        module._parameters[tensor_name] = tied_params_map[old_value.data_ptr()][device]
        return
    if old_value.device == torch.device("meta") and device not in ["meta", torch.device("meta")] and value is None: raise ValueError(f"{tensor_name} is on the meta device, we need a `value` to put in on {device}.")
    param = module._parameters[tensor_name] if tensor_name in module._parameters else None
    param_cls = type(param)
    if value is not None:
        if old_value.shape != value.shape and param_cls.__name__ != "Params4bit": raise ValueError(f'Trying to set a tensor of shape {value.shape} in "{tensor_name}" (which has shape {old_value.shape}), this looks incorrect.')
        if dtype is None: value = value.to(old_value.dtype)
        elif not str(value.dtype).startswith(("torch.uint", "torch.int", "torch.bool")): value = value.to(dtype)
    device_quantization = None
    with torch.no_grad():
        if (param is not None and param.device.type != "cuda" and torch.device(device).type == "cuda" and param_cls.__name__ in ["Int8Params", "FP4Params", "Params4bit"]):
            device_quantization = device
            device = "cpu"
        if isinstance(device, int):
            if is_npu_available(): device = f"npu:{device}"
            elif is_mlu_available(): device = f"mlu:{device}"
            elif is_musa_available(): device = f"musa:{device}"
            elif is_xpu_available(): device = f"xpu:{device}"
        if "xpu" in str(device) and not is_xpu_available(): raise ValueError(f'{device} is not available, you should use device="cpu" instead')
        if value is None:
            new_value = old_value.to(device)
            if dtype is not None and device in ["meta", torch.device("meta")]:
                if not str(old_value.dtype).startswith(("torch.uint", "torch.int", "torch.bool")): new_value = new_value.to(dtype)
                if not is_buffer: module._parameters[tensor_name] = param_cls(new_value, requires_grad=old_value.requires_grad)
        elif isinstance(value, torch.Tensor): new_value = value.to(device)
        else: new_value = torch.tensor(value, device=device)
        if device_quantization is not None: device = device_quantization
        if is_buffer: module._buffers[tensor_name] = new_value
        elif value is not None or not check_device_same(torch.device(device), module._parameters[tensor_name].device):
            param_cls = type(module._parameters[tensor_name])
            kwargs = module._parameters[tensor_name].__dict__
            if param_cls.__name__ in ["Int8Params", "FP4Params", "Params4bit"]:
                if param_cls.__name__ == "Int8Params" and new_value.dtype == torch.float32: new_value = new_value.to(torch.float16)
                if device == "cpu" and param_cls.__name__ == "Int8Params":
                    new_value = param_cls(new_value, requires_grad=old_value.requires_grad, **kwargs).to(0).to("cpu")
                    new_value.CB = new_value.CB.to("cpu")
                    new_value.SCB = new_value.SCB.to("cpu")
                else: new_value = param_cls(new_value, requires_grad=old_value.requires_grad, **kwargs).to(device)
            elif param_cls.__name__ in ["QTensor", "QBitsTensor"]: new_value = torch.nn.Parameter(new_value, requires_grad=old_value.requires_grad).to(device)
            elif param_cls.__name__ in ["AffineQuantizedTensor"]: new_value = torch.nn.Parameter(param_cls(new_value.layout_tensor, new_value.block_size,
            new_value.shape, new_value.quant_min, new_value.quant_max, new_value.zero_point_domain), requires_grad=old_value.requires_grad).to(device)
            else: new_value = param_cls(new_value, requires_grad=old_value.requires_grad).to(device)
            module._parameters[tensor_name] = new_value
            if fp16_statistics is not None:
                module._parameters[tensor_name].SCB = fp16_statistics.to(device)
                del fp16_statistics
            if (module.__class__.__name__ == "Linear8bitLt" and getattr(module.weight, "SCB", None) is None and str(module.weight.device) != "meta"):
                device_index = torch.device(device).index if torch.device(device).type == "cuda" else None
                if not getattr(module.weight, "SCB", None) and device_index is not None:
                    if module.bias is not None and module.bias.device.type != "meta": module = module.cuda(device_index)
                    elif module.bias is None: module = module.cuda(device_index)
            elif (module.__class__.__name__ == "Linear4bit" and getattr(module.weight, "quant_state", None) is None and str(module.weight.device) != "meta"):
                device_index = torch.device(device).index if torch.device(device).type == "cuda" else None
                if not getattr(module.weight, "quant_state", None) and device_index is not None: module.weight = module.weight.cuda(device_index)
    if device != "cpu": clear_device_cache()
    if (tied_params_map is not None and old_value.data_ptr() in tied_params_map and device not in tied_params_map[old_value.data_ptr()]): tied_params_map[old_value.data_ptr()][device] = new_value
    elif (value is not None and tied_params_map is not None and value.data_ptr() in tied_params_map and device not in tied_params_map[value.data_ptr()]): tied_params_map[value.data_ptr()][device] = new_value
def named_module_tensors(module: nn.Module, include_buffers: bool = True, recurse: bool = False, remove_non_persistent: bool = False):
    yield from module.named_parameters(recurse=recurse)
    if include_buffers:
        non_persistent_buffers = set()
        if remove_non_persistent: non_persistent_buffers = get_non_persistent_buffers(module, recurse=recurse)
        for named_buffer in module.named_buffers(recurse=recurse):
            name, _ = named_buffer
            if name not in non_persistent_buffers: yield named_buffer
def get_non_persistent_buffers(module: nn.Module, recurse: bool = False):
    non_persistent_buffers_set = module._non_persistent_buffers_set
    if recurse:
        for _, m in module.named_modules(): non_persistent_buffers_set |= m._non_persistent_buffers_set
    return non_persistent_buffers_set
class FindTiedParametersResult(list):
    def __init__(self, *args, **kwargs): super().__init__(*args, **kwargs)
    def values(self): return sum([x[1:] for x in self], [])
def check_tied_parameters_in_config(model: nn.Module):
    has_tied_word_embedding = False
    has_tied_encoder_decoder = False
    has_tied_module = False
    if "PreTrainedModel" in [c.__name__ for c in inspect.getmro(model.__class__)]:
        has_tied_word_embedding = (hasattr(model, "config") and getattr(model.config, "tie_word_embeddings", False) and model.get_output_embeddings())
        has_tied_encoder_decoder = (hasattr(model, "config") and getattr(model.config, "is_encoder_decoder", False) and getattr(model.config, "tie_encoder_decoder", False))
        has_tied_module = any(hasattr(module, "_tie_weights") for module in model.modules())
    return any([has_tied_word_embedding, has_tied_encoder_decoder, has_tied_module])
def _get_param_device(param, device_map):
    if param in device_map: return device_map[param]
    parent_param = ".".join(param.split(".")[:-1])
    if parent_param == param: raise ValueError(f"The `device_map` does not contain the module {param}.")
    else: return _get_param_device(parent_param, device_map)
def check_tied_parameters_on_same_device(tied_params, device_map):
    for tie_param in tied_params:
        tie_param_devices = {}
        for param in tie_param: tie_param_devices[param] = _get_param_device(param, device_map)
def _get_named_modules(module: torch.nn.Module, memo: Optional[Set[torch.nn.Module]] = None, prefix: str = "", remove_duplicate: bool = True):
    if memo is None: memo = set()
    if module not in memo:
        if remove_duplicate: memo.add(module)
        yield prefix, module
        for name, sub_module in module._modules.items():
            if sub_module is None: continue
            submodule_prefix = prefix + ("." if prefix else "") + name
            yield from _get_named_modules(sub_module, memo, submodule_prefix, remove_duplicate)
def _get_named_parameters(module: torch.nn.Module, prefix="", recurse=True, remove_duplicate: bool = True):
    memo = set()
    modules = (_get_named_modules(module, prefix=prefix, remove_duplicate=remove_duplicate) if recurse else [(prefix, module)])
    for module_prefix, module in modules:
        members = module._parameters.items()
        for k, v in members:
            if v is None or v in memo: continue
            if remove_duplicate: memo.add(v)
            name = module_prefix + ("." if module_prefix else "") + k
            yield name, v
def find_tied_parameters(model: torch.nn.Module, **kwargs):
    all_named_parameters = {name: param for name, param in _get_named_parameters(model, remove_duplicate=False)}
    no_duplicate_named_parameters = {name: param for name, param in _get_named_parameters(model, remove_duplicate=True)}
    tied_param_names = set(all_named_parameters.keys()) - set(no_duplicate_named_parameters.keys())
    tied_param_groups = {}
    for tied_param_name in tied_param_names:
        tied_param = all_named_parameters[tied_param_name]
        for param_name, param in no_duplicate_named_parameters.items():
            if param is tied_param:
                if param_name not in tied_param_groups: tied_param_groups[param_name] = []
                tied_param_groups[param_name].append(tied_param_name)
    return FindTiedParametersResult([sorted([weight] + list(set(tied))) for weight, tied in tied_param_groups.items()])
def retie_parameters(model, tied_params):
    for tied_group in tied_params:
        param_to_tie = None
        for param_name in tied_group:
            module = model
            splits = param_name.split(".")
            for split in splits[:-1]: module = getattr(module, split)
            param = getattr(module, splits[-1])
            if param_to_tie is None and param.device != torch.device("meta"):
                param_to_tie = param
                break
        if param_to_tie is not None:
            for param_name in tied_group:
                module = model
                splits = param_name.split(".")
                for split in splits[:-1]: module = getattr(module, split)
                setattr(module, splits[-1], param_to_tie)
def _get_proper_dtype(dtype: Union[str, torch.device]) -> torch.dtype:
    if isinstance(dtype, str):
        dtype = dtype.replace("torch.", "")
        dtype = getattr(torch, dtype)
    return dtype
def compute_module_sizes(model: nn.Module, dtype: Optional[Union[str, torch.device]] = None, special_dtypes: Optional[Dict[str, Union[str, torch.device]]] = None, buffers_only: bool = False):
    if dtype is not None:
        dtype = _get_proper_dtype(dtype)
        dtype_size = dtype_byte_size(dtype)
    if special_dtypes is not None:
        special_dtypes = {key: _get_proper_dtype(dtyp) for key, dtyp in special_dtypes.items()}
        special_dtypes_size = {key: dtype_byte_size(dtyp) for key, dtyp in special_dtypes.items()}
    module_sizes = defaultdict(int)
    module_list = []
    if not buffers_only: module_list = named_module_tensors(model, recurse=True)
    else: module_list = model.named_buffers(recurse=True)
    for name, tensor in module_list:
        if special_dtypes is not None and name in special_dtypes: size = tensor.numel() * special_dtypes_size[name]
        elif dtype is None: size = tensor.numel() * dtype_byte_size(tensor.dtype)
        elif str(tensor.dtype).startswith(("torch.uint", "torch.int", "torch.bool")): size = tensor.numel() * dtype_byte_size(tensor.dtype)
        else: size = tensor.numel() * min(dtype_size, dtype_byte_size(tensor.dtype))
        name_parts = name.split(".")
        for idx in range(len(name_parts) + 1): module_sizes[".".join(name_parts[:idx])] += size
    return module_sizes
def compute_module_total_buffer_size(model: nn.Module, dtype: Optional[Union[str, torch.device]] = None, special_dtypes: Optional[Dict[str, Union[str, torch.device]]] = None):
    module_sizes = compute_module_sizes(model, dtype=dtype, special_dtypes=special_dtypes, buffers_only=True)
    return module_sizes.get("", 0)
def get_max_layer_size(modules: List[Tuple[str, torch.nn.Module]], module_sizes: Dict[str, int], no_split_module_classes: List[str]):
    max_size = 0
    layer_names = []
    modules_to_treat = modules.copy()
    while len(modules_to_treat) > 0:
        module_name, module = modules_to_treat.pop(0)
        modules_children = list(module.named_children()) if isinstance(module, torch.nn.Module) else []
        if len(modules_children) == 0 or module.__class__.__name__ in no_split_module_classes:
            size = module_sizes[module_name]
            if size > max_size:
                max_size = size
                layer_names = [module_name]
            elif size == max_size: layer_names.append(module_name)
        else: modules_to_treat = [(f"{module_name}.{n}", v) for n, v in modules_children] + modules_to_treat
    return max_size, layer_names
def get_max_memory(max_memory: Optional[Dict[Union[int, str], Union[int, str]]] = None):
    import psutil
    if max_memory is None:
        max_memory = {}
        if is_npu_available():
            for i in range(torch.npu.device_count()):
                try:
                    _ = torch.tensor(0, device=torch.device("npu", i))
                    max_memory[i] = torch.npu.mem_get_info(i)[0]
                except Exception: continue
        elif is_mlu_available():
            for i in range(torch.mlu.device_count()):
                try:
                    _ = torch.tensor(0, device=torch.device("mlu", i))
                    max_memory[i] = torch.mlu.mem_get_info(i)[0]
                except Exception: continue
        elif is_musa_available():
            for i in range(torch.musa.device_count()):
                try:
                    _ = torch.tensor(0, device=torch.device("musa", i))
                    max_memory[i] = torch.musa.mem_get_info(i)[0]
                except Exception: continue
        elif is_xpu_available():
            for i in range(torch.xpu.device_count()):
                try:
                    _ = torch.tensor(0, device=torch.device("xpu", i))
                    max_memory[i] = get_xpu_available_memory(i)
                except Exception: continue
        else:
            for i in range(torch.cuda.device_count()):
                try:
                    _ = torch.tensor([0], device=i)
                    max_memory[i] = torch.cuda.mem_get_info(i)[0]
                except Exception: continue
        if is_mps_available(): max_memory["mps"] = psutil.virtual_memory().available
        else: max_memory["cpu"] = psutil.virtual_memory().available
        return max_memory
    for key in max_memory:
        if isinstance(max_memory[key], str): max_memory[key] = convert_file_size_to_int(max_memory[key])
    gpu_devices = [k for k in max_memory.keys() if isinstance(k, int)]
    gpu_devices.sort()
    if is_npu_available(): num_devices = torch.npu.device_count()
    elif is_mlu_available(): num_devices = torch.mlu.device_count()
    elif is_musa_available(): num_devices = torch.musa.device_count()
    elif is_xpu_available(): num_devices = torch.xpu.device_count()
    else: num_devices = torch.cuda.device_count()
    all_devices = gpu_devices + [k for k in ["mps", "cpu", "disk"] if k in max_memory.keys()]
    for k in max_memory.keys():
        if k not in all_devices: raise ValueError(f"Device {k} is not recognized, available devices are integers(for GPU/XPU), 'mps', 'cpu' and 'disk'")
    max_memory = {k: max_memory[k] for k in all_devices}
    return max_memory
def clean_device_map(device_map: Dict[str, Union[int, str, torch.device]], module_name: str = ""):
    prefix = "" if module_name == "" else f"{module_name}."
    values = [v for k, v in device_map.items() if k.startswith(prefix)]
    if len(set(values)) == 1 and len(values) > 1:
        for k in [k for k in device_map if k.startswith(prefix)]: del device_map[k]
        device_map[module_name] = values[0]
    children_modules = [k for k in device_map.keys() if k.startswith(prefix) and len(k) > len(module_name)]
    idx = len(module_name.split(".")) + 1 if len(module_name) > 0 else 1
    children_modules = set(".".join(k.split(".")[:idx]) for k in children_modules)
    for child in children_modules: clean_device_map(device_map, module_name=child)
    return device_map
def load_offloaded_weights(model, index, offload_folder):
    if index is None or len(index) == 0: return
    for param_name, metadata in index.items():
        if "SCB" in param_name: continue
        fp16_statistics = None
        if "weight" in param_name and param_name.replace("weight", "SCB") in index.keys():
            weight_name = param_name.replace("weight", "SCB")
            fp16_statistics = load_offloaded_weight(os.path.join(offload_folder, f"{weight_name}.dat"), index[weight_name])
        tensor_file = os.path.join(offload_folder, f"{param_name}.dat")
        weight = load_offloaded_weight(tensor_file, metadata)
        set_module_tensor_to_device(model, param_name, "cpu", value=weight, fp16_statistics=fp16_statistics)
def get_module_leaves(module_sizes):
    module_children = {}
    for module in module_sizes:
        if module == "" or "." not in module: continue
        parent = module.rsplit(".", 1)[0]
        module_children[parent] = module_children.get(parent, 0) + 1
    leaves = [module for module in module_sizes if module_children.get(module, 0) == 0 and module != ""]
    return leaves
def get_balanced_memory(model: nn.Module, max_memory: Optional[Dict[Union[int, str], Union[int, str]]] = None, no_split_module_classes: Optional[List[str]] = None,
dtype: Optional[Union[str, torch.dtype]] = None, special_dtypes: Optional[Dict[str, Union[str, torch.device]]] = None, low_zero: bool = False):
    user_not_set_max_memory = max_memory is None
    max_memory = get_max_memory(max_memory)
    if is_npu_available(): expected_device_type = "npu"
    elif is_mlu_available(): expected_device_type = "mlu"
    elif is_musa_available(): expected_device_type = "musa"
    elif is_xpu_available(): expected_device_type = "xpu"
    else: expected_device_type = "cuda"
    num_devices = len([d for d in max_memory if torch.device(d).type == expected_device_type and max_memory[d] > 0])
    if num_devices == 0: return max_memory
    if num_devices == 1:
        low_zero = False
        if user_not_set_max_memory:
            for key in max_memory.keys():
                if isinstance(key, int):
                    max_memory[key] *= 0.9
                    break
    module_sizes = compute_module_sizes(model, dtype=dtype, special_dtypes=special_dtypes)
    per_gpu = module_sizes[""] // (num_devices - 1 if low_zero else num_devices)
    if no_split_module_classes is None: no_split_module_classes = []
    elif not isinstance(no_split_module_classes, (list, tuple)): no_split_module_classes = [no_split_module_classes]
    if len(no_split_module_classes) > 0:
        no_split_children = {}
        for name, size in module_sizes.items():
            if name == "": continue
            submodule = model
            for submodule_name in name.split("."): submodule = getattr(submodule, submodule_name)
            class_name = submodule.__class__.__name__
            if class_name in no_split_module_classes and class_name not in no_split_children: no_split_children[class_name] = size
            if set(no_split_children.keys()) == set(no_split_module_classes): break
        buffer = max(no_split_children.values()) if len(no_split_children) > 0 else 0
    else: buffer = 0
    leaves = get_module_leaves(module_sizes)
    module_sizes = {n: v for n, v in module_sizes.items() if n not in leaves}
    leaves = get_module_leaves(module_sizes)
    mean_leaves = int(sum([module_sizes[n] for n in leaves]) / max(len(leaves), 1))
    buffer = int(1.25 * max(buffer, mean_leaves))
    per_gpu += buffer
    gpus_idx_list = list(sorted(device_id for device_id, device_mem in max_memory.items() if isinstance(device_id, int) and device_mem > 0))
    for idx in gpus_idx_list[:-1]: max_memory[idx] = min(max_memory[0] if low_zero and idx == 0 else per_gpu, max_memory[idx])
    if low_zero:
        min_zero = max(0, module_sizes[""] - sum([max_memory[i] for i in range(1, num_devices)]))
        max_memory[0] = min(min_zero, max_memory[0])
    return max_memory
def calculate_maximum_sizes(model: torch.nn.Module):
    sizes = compute_module_sizes(model)
    no_split_modules = getattr(model, "_no_split_modules", None)
    if no_split_modules is None: no_split_modules = []
    modules_to_treat = (list(model.named_parameters(recurse=False)) + list(model.named_children()) + list(model.named_buffers(recurse=False)))
    largest_layer = get_max_layer_size(modules_to_treat, sizes, no_split_modules)
    total_size = sizes[""]
    return total_size, largest_layer
def infer_auto_device_map(model: nn.Module, max_memory: Optional[Dict[Union[int, str], Union[int, str]]] = None, no_split_module_classes: Optional[List[str]] = None,
dtype: Optional[Union[str, torch.dtype]] = None, special_dtypes: Optional[Dict[str, Union[str, torch.dtype]]] = None, verbose: bool = False,
clean_result: bool = True, offload_buffers: bool = False):
    max_memory = get_max_memory(max_memory)
    if no_split_module_classes is None: no_split_module_classes = []
    elif not isinstance(no_split_module_classes, (list, tuple)): no_split_module_classes = [no_split_module_classes]
    devices = list(max_memory.keys())
    if "disk" not in devices: devices.append("disk")
    gpus = [device for device in devices if device not in ["cpu", "disk"]]
    if "mps" in gpus: main_devices = ["mps"]
    elif len(gpus) > 0: main_devices = [gpus[0], "cpu"]
    else: main_devices = ["cpu"]
    module_sizes = compute_module_sizes(model, dtype=dtype, special_dtypes=special_dtypes)
    tied_parameters = find_tied_parameters(model)
    device_map = OrderedDict()
    current_device = 0
    current_memory_used = 0
    device_memory_used = {}
    device_buffer_sizes = {}
    modules_to_treat = (list(model.named_parameters(recurse=False)) + list(model.named_children()) + list(model.named_buffers(recurse=False)))
    max_layer_size, max_layer_names = get_max_layer_size(modules_to_treat, module_sizes, no_split_module_classes)
    while len(modules_to_treat) > 0:
        name, module = modules_to_treat.pop(0)
        max_layer_names = [n for n in max_layer_names if n != name and not n.startswith(name + ".")]
        if len(max_layer_names) == 0: max_layer_size, max_layer_names = get_max_layer_size([(n, m) for n, m in modules_to_treat if isinstance(m, torch.nn.Module)], module_sizes, no_split_module_classes)
        module_size = module_sizes[name]
        tied_param_goups = [tied_group for tied_group in tied_parameters if any(name + "." in k + "." for k in tied_group) and not all(name + "." in k + "." for k in tied_group)]
        tied_params = sum([[p for p in tied_group if name + "." not in p + "."] for tied_group in tied_param_goups], [])
        device = devices[current_device]
        current_max_size = max_memory[device] if device != "disk" else None
        current_memory_reserved = 0
        if devices[current_device] in main_devices:
            current_max_size = current_max_size - max_layer_size
            current_memory_reserved = max_layer_size
        if current_max_size is not None and current_memory_used + module_size > current_max_size:
            modules_children = ([] if isinstance(module, nn.Parameter) or isinstance(module, torch.Tensor) else list(module.named_children()))
            if len(modules_children) == 0 or module.__class__.__name__ in no_split_module_classes:
                device_memory_used[device] = current_memory_used + current_memory_reserved
                current_device += 1
                modules_to_treat = [(name, module)] + modules_to_treat
                current_memory_used = 0
            else:
                modules_children = list(module.named_parameters(recurse=False)) + modules_children
                modules_to_treat = [(f"{name}.{n}", v) for n, v in modules_children] + modules_to_treat
                max_layer_size, max_layer_names = get_max_layer_size([(n, m) for n, m in modules_to_treat if isinstance(m, torch.nn.Module)], module_sizes, no_split_module_classes)
        elif len(tied_params) > 0:
            tied_module_names = []
            tied_modules = []
            for tied_param in tied_params:
                tied_module_index = [i for i, (n, _) in enumerate(modules_to_treat) if n in tied_param][0]
                tied_module_names.append(modules_to_treat[tied_module_index][0])
                tied_modules.append(modules_to_treat[tied_module_index][1])
            module_size_with_ties = module_size
            for tied_param, tied_module_name in zip(tied_params, tied_module_names): module_size_with_ties += module_sizes[tied_module_name] - module_sizes[tied_param]
            if current_max_size is None or current_memory_used + module_size_with_ties <= current_max_size:
                current_memory_used += module_size_with_ties
                device_map[name] = devices[current_device]
                for tied_module_name in tied_module_names:
                    if tied_module_name in [m[0] for m in modules_to_treat]:
                        tied_module_index = [i for i, (n, _) in enumerate(modules_to_treat) if n == tied_module_name][0]
                        modules_to_treat.pop(tied_module_index)
                    device_map[tied_module_name] = devices[current_device]
                if not offload_buffers and isinstance(module, nn.Module):
                    current_buffer_size = compute_module_total_buffer_size(module, dtype=dtype, special_dtypes=special_dtypes)
                    device_buffer_sizes[device] = device_buffer_sizes.get(device, 0) + current_buffer_size
            else:
                split_happened = False
                for tied_module_name, tied_module in zip(tied_module_names, tied_modules):
                    tied_module_children = list(tied_module.named_children())
                    if len(tied_module_children) == 0 or tied_module.__class__.__name__ in no_split_module_classes: continue
                    tied_module_children = list(tied_module.named_parameters(recurse=False)) + tied_module_children
                    tied_module_children = [(f"{tied_module_name}.{n}", v) for n, v in tied_module_children]
                    tied_module_index = [i for i, (n, _) in enumerate(modules_to_treat) if n == tied_module_name][0]
                    modules_to_treat = ([(name, module)] + modules_to_treat[:tied_module_index] + tied_module_children + modules_to_treat[tied_module_index + 1 :])
                    max_layer_size, max_layer_names = get_max_layer_size([(n, m) for n, m in modules_to_treat if isinstance(m, torch.nn.Module)], module_sizes, no_split_module_classes)
                    split_happened = True
                    break
                if not split_happened:
                    device_memory_used[device] = current_memory_used + current_memory_reserved
                    current_device += 1
                    modules_to_treat = [(name, module)] + modules_to_treat
                    current_memory_used = 0
        else:
            current_memory_used += module_size
            device_memory_used[device] = current_memory_used + current_memory_reserved
            device_map[name] = devices[current_device]
            if not offload_buffers and isinstance(module, nn.Module):
                current_buffer_size = compute_module_total_buffer_size(module, dtype=dtype, special_dtypes=special_dtypes)
                device_buffer_sizes[device] = device_buffer_sizes.get(device, 0) + current_buffer_size
    if clean_result: device_map = clean_device_map(device_map)
    non_gpu_buffer_size = device_buffer_sizes.get("cpu", 0) + device_buffer_sizes.get("disk", 0)
    if non_gpu_buffer_size > 0 and not offload_buffers:
        is_buffer_fit_any_gpu = False
        for gpu_device, gpu_max_memory in max_memory.items():
            if gpu_device == "cpu" or gpu_device == "disk": continue
            if not is_buffer_fit_any_gpu:
                gpu_memory_used = device_memory_used.get(gpu_device, 0)
                if gpu_max_memory >= non_gpu_buffer_size + gpu_memory_used: is_buffer_fit_any_gpu = True
    return device_map
def check_device_map(model: nn.Module, device_map: Dict[str, Union[int, str, torch.device]]):
    all_model_tensors = [name for name, _ in model.state_dict().items()]
    for module_name in device_map.keys():
        if module_name == "":
            all_model_tensors.clear()
            break
        else: all_model_tensors = [name for name in all_model_tensors if not name == module_name and not name.startswith(module_name + ".")]
    if len(all_model_tensors) > 0:
        non_covered_params = ", ".join(all_model_tensors)
        raise ValueError(f"The device_map provided does not give any device for the following parameters: {non_covered_params}")
def load_state_dict(checkpoint_file, device_map=None):
    if checkpoint_file.endswith(".safetensors"):
        with safe_open(checkpoint_file, framework="pt") as f:
            metadata = f.metadata()
            weight_names = f.keys()
        if metadata is None: metadata = {"format": "pt"}
        if metadata.get("format") not in ["pt", "tf", "flax"]: raise OSError(f"The safetensors archive passed at {checkpoint_file} does not contain the valid metadata. Make sure you save your model with the `save_pretrained` method.")
        elif metadata["format"] != "pt": raise ValueError(f"The checkpoint passed was saved with {metadata['format']}, we need a the pt format.")
        if device_map is None: return safe_load_file(checkpoint_file)
        else:
            if len(set(device_map.values())) == 1:
                device = list(device_map.values())[0]
                target_device = device
                if is_xpu_available():
                    if compare_versions("safetensors", "<", "0.4.2"): raise ImportError("Safetensors version must be >= 0.4.2 for XPU. Please upgrade safetensors.")
                    if isinstance(device, int): target_device = f"xpu:{device}"
                return safe_load_file(checkpoint_file, device=target_device)
            devices = list(set(device_map.values()) - {"disk"})
            if "cpu" not in devices: devices.append("cpu")
            device_weights = {device: [] for device in devices}
            for module_name, device in device_map.items():
                if device in devices: device_weights[device].extend([k for k in weight_names if k == module_name or k.startswith(module_name + ".")])
            device_weights["cpu"].extend([k for k in weight_names if k not in sum(device_weights.values(), [])])
            tensors = {}
            if is_tqdm_available(): progress_bar = tqdm(main_process_only=False, total=sum([len(device_weights[device]) for device in devices]), unit="w", smoothing=0, leave=False)
            else: progress_bar = None
            for device in devices:
                target_device = device
                if is_xpu_available():
                    if compare_versions("safetensors", "<", "0.4.2"): raise ImportError("Safetensors version must be >= 0.4.2 for XPU. Please upgrade safetensors.")
                    if isinstance(device, int): target_device = f"xpu:{device}"
                with safe_open(checkpoint_file, framework="pt", device=target_device) as f:
                    for key in device_weights[device]:
                        if progress_bar is not None:
                            progress_bar.set_postfix(dev=device, refresh=False)
                            progress_bar.set_description(key)
                        tensors[key] = f.get_tensor(key)
                        if progress_bar is not None: progress_bar.update()
            if progress_bar is not None: progress_bar.close()
            return tensors
    else: return torch.load(checkpoint_file, map_location=torch.device("cpu"))
def get_state_dict_offloaded_model(model: nn.Module):
    from ..hooks import AlignDevicesHook
    state_dict = {}
    placeholders = set()
    for name, module in model.named_modules():
        if name == "": continue
        if hasattr(module, "_hf_hook") and isinstance(module._hf_hook, AlignDevicesHook) and module._hf_hook.offload:
            original_device = module._hf_hook.execution_device
            module._hf_hook.execution_device = "cpu"
            try: module._hf_hook.pre_forward(module)
            except MemoryError: raise MemoryError("Offloaded module must fit in CPU memory to call save_model!") from None
            module_state_dict = module.state_dict()
            module._hf_hook.post_forward(module, torch.tensor([]))
            module._hf_hook.execution_device = original_device
        else: module_state_dict = module.state_dict()
        for key in module_state_dict:
            if module_state_dict[key].device == torch.device("meta"):
                placeholders.add(name + f".{key}")
                continue
            params = module_state_dict[key]
            state_dict[name + f".{key}"] = params
    for key in placeholders.copy():
        if key in state_dict: placeholders.remove(key)
    return state_dict
def get_state_dict_from_offload(module: nn.Module, module_name: str, state_dict: Dict[str, Union[str, torch.tensor]], device_to_put_offload: Union[int, str, torch.device] = "cpu"):
    from ..hooks import AlignDevicesHook
    root = module_name[: module_name.rfind(".")]
    preforward = False
    if hasattr(module, "_hf_hook") and isinstance(module._hf_hook, AlignDevicesHook) and module._hf_hook.offload:
        original_device = module._hf_hook.execution_device
        module._hf_hook.execution_device = device_to_put_offload
        module._hf_hook.pre_forward(module)
        preforward = True
    for m_key in module.state_dict():
        params = module.state_dict()[m_key]
        if (root + f".{m_key}") in state_dict: state_dict[root + f".{m_key}"] = params
    if preforward:
        module._hf_hook.post_forward(module, torch.tensor([]))
        module._hf_hook.execution_device = original_device
    return state_dict
def load_checkpoint_in_model(model: nn.Module, checkpoint: Union[str, os.PathLike], device_map: Optional[Dict[str, Union[int, str, torch.device]]] = None,
offload_folder: Optional[Union[str, os.PathLike]] = None, dtype: Optional[Union[str, torch.dtype]] = None, offload_state_dict: bool = False, offload_buffers: bool = False,
keep_in_fp32_modules: List[str] = None, offload_8bit_sapiens: bool = False, strict: bool = False):
    if offload_8bit_sapiens: from .sapiens import quantize_and_offload_8bit
    tied_params = find_tied_parameters(model)
    if device_map is not None: check_tied_parameters_on_same_device(tied_params, device_map)
    if offload_folder is None and device_map is not None and "disk" in device_map.values(): raise ValueError("At least one of the model submodule will be offloaded to disk, please pass along an `offload_folder`.")
    elif offload_folder is not None and device_map is not None and "disk" in device_map.values(): os.makedirs(offload_folder, exist_ok=True)
    if isinstance(dtype, str):
        dtype = dtype.replace("torch.", "")
        dtype = getattr(torch, dtype)
    checkpoint_files = None
    index_filename = None
    if os.path.isfile(checkpoint):
        if str(checkpoint).endswith(".json"): index_filename = checkpoint
        else: checkpoint_files = [checkpoint]
    elif os.path.isdir(checkpoint):
        potential_state_bin = [f for f in os.listdir(checkpoint) if f == WEIGHTS_NAME]
        potential_state_safetensor = [f for f in os.listdir(checkpoint) if f == SAFE_WEIGHTS_NAME]
        if len(potential_state_bin) == 1: checkpoint_files = [os.path.join(checkpoint, potential_state_bin[0])]
        elif len(potential_state_safetensor) == 1: checkpoint_files = [os.path.join(checkpoint, potential_state_safetensor[0])]
        else:
            potential_index = [f for f in os.listdir(checkpoint) if f.endswith(".index.json")]
            if len(potential_index) == 0: raise ValueError(f"{checkpoint} is not a folder containing a `.index.json` file or a {WEIGHTS_NAME} or a {SAFE_WEIGHTS_NAME} file")
            elif len(potential_index) == 1: index_filename = os.path.join(checkpoint, potential_index[0])
            else: raise ValueError(f"{checkpoint} containing more than one `.index.json` file, delete the irrelevant ones.")
    else: raise ValueError(f"`checkpoint` should be the path to a file containing a whole state dict, or the index of a sharded checkpoint, or a folder containing a sharded checkpoint or the whole state dict, but got {checkpoint}.")
    if index_filename is not None:
        checkpoint_folder = os.path.split(index_filename)[0]
        with open(index_filename) as f: index = json.loads(f.read())
        if "weight_map" in index: index = index["weight_map"]
        checkpoint_files = sorted(list(set(index.values())))
        checkpoint_files = [os.path.join(checkpoint_folder, f) for f in checkpoint_files]
    offload_index = {}
    if offload_state_dict:
        state_dict_folder = tempfile.mkdtemp()
        state_dict_index = {}
    unexpected_keys = set()
    model_keys = set(model.state_dict().keys())
    buffer_names = [name for name, _ in model.named_buffers()]
    for checkpoint_file in checkpoint_files:
        loaded_checkpoint = load_state_dict(checkpoint_file, device_map=device_map)
        if device_map is None:
            model.load_state_dict(loaded_checkpoint, strict=strict)
            unexpected_keys.update(set(loaded_checkpoint.keys()) - model_keys)
        else:
            for param_name, param in loaded_checkpoint.items():
                if "SCB" in param_name: continue
                if param_name not in model_keys:
                    unexpected_keys.add(param_name)
                    if not strict: continue
                module_name = param_name
                while len(module_name) > 0 and module_name not in device_map: module_name = ".".join(module_name.split(".")[:-1])
                if module_name == "" and "" not in device_map: raise ValueError(f"{param_name} doesn't have any device set.")
                param_device = device_map[module_name]
                new_dtype = dtype
                if dtype is not None and torch.is_floating_point(param):
                    if keep_in_fp32_modules is not None and dtype == torch.float16:
                        proceed = False
                        for key in keep_in_fp32_modules:
                            if ((key in param_name) and (key + "." in param_name)) or key == param_name:
                                proceed = True
                                break
                        if proceed: new_dtype = torch.float32
                if "weight" in param_name and param_name.replace("weight", "SCB") in loaded_checkpoint.keys():
                    if param.dtype == torch.int8: fp16_statistics = loaded_checkpoint[param_name.replace("weight", "SCB")]
                else: fp16_statistics = None
                if param_device == "disk":
                    if offload_buffers or param_name not in buffer_names:
                        if new_dtype is None: new_dtype = param.dtype
                        if offload_8bit_sapiens:
                            quantize_and_offload_8bit(model, param, param_name, new_dtype, offload_folder, offload_index, fp16_statistics)
                            continue
                        else: set_module_tensor_to_device(model, param_name, "meta", dtype=new_dtype)
                        offload_weight(param, param_name, offload_folder, index=offload_index)
                elif param_device == "cpu" and offload_state_dict:
                    if new_dtype is None: new_dtype = param.dtype
                    if offload_8bit_sapiens: quantize_and_offload_8bit(model, param, param_name, new_dtype, state_dict_folder, state_dict_index, fp16_statistics)
                    else:
                        set_module_tensor_to_device(model, param_name, "meta", dtype=new_dtype)
                        offload_weight(param, param_name, state_dict_folder, index=state_dict_index)
                else: set_module_tensor_to_device(model, param_name, param_device, value=param, dtype=new_dtype, fp16_statistics=fp16_statistics)
        del loaded_checkpoint
        gc.collect()
    save_offload_index(offload_index, offload_folder)
    if offload_state_dict:
        load_offloaded_weights(model, state_dict_index, state_dict_folder)
        shutil.rmtree(state_dict_folder)
    retie_parameters(model, tied_params)
def get_mixed_precision_context_manager(native_amp: bool = False, autocast_kwargs: AutocastKwargs = None):
    state = AcceleratorState()
    if autocast_kwargs is None: autocast_kwargs = {}
    else: autocast_kwargs = autocast_kwargs.to_kwargs()
    if native_amp:
        device_type = ("cuda" if (state.distributed_type == DistributedType.XLA and is_torch_xla_available(check_is_gpu=True)) else state.device.type)
        if state.mixed_precision == "fp16": return torch.autocast(device_type=device_type, dtype=torch.float16, **autocast_kwargs)
        elif state.mixed_precision in ["bf16", "fp8"] and state.distributed_type in [DistributedType.NO, DistributedType.MULTI_CPU, DistributedType.MULTI_GPU,
        DistributedType.MULTI_MLU, DistributedType.MULTI_MUSA, DistributedType.MULTI_NPU, DistributedType.MULTI_XPU, DistributedType.FSDP, DistributedType.XLA]: return torch.autocast(device_type=device_type, dtype=torch.bfloat16, **autocast_kwargs)
        else: return torch.autocast(device_type=device_type, **autocast_kwargs)
    else: return contextlib.nullcontext()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import builtins
import sys
from ...utils.imports import _is_package_available
from . import cursor, input
from .helpers import Direction, clear_line, forceWrite, linebreak, move_cursor, reset_cursor, writeColor
from .keymap import KEYMAP
in_colab = False
try: in_colab = _is_package_available("google.colab")
except ModuleNotFoundError: pass
@input.register
class BulletMenu:
    def __init__(self, prompt: str = None, choices: list = []):
        self.position = 0
        self.choices = choices
        self.prompt = prompt
        if sys.platform == "win32": self.arrow_char = "*"
        else: self.arrow_char = "➔ "
    def write_choice(self, index, end: str = ""):
        if sys.platform != "win32": writeColor(self.choices[index], 32, end)
        else: forceWrite(self.choices[index], end)
    def print_choice(self, index: int):
        "Prints the choice at the given index"
        if index == self.position:
            forceWrite(f" {self.arrow_char} ")
            self.write_choice(index)
        else: forceWrite(f"    {self.choices[index]}")
        reset_cursor()
    def move_direction(self, direction: Direction, num_spaces: int = 1):
        old_position = self.position
        if direction == Direction.DOWN:
            if self.position + 1 >= len(self.choices): return
            self.position += num_spaces
        else:
            if self.position - 1 < 0: return
            self.position -= num_spaces
        clear_line()
        self.print_choice(old_position)
        move_cursor(num_spaces, direction.name)
        self.print_choice(self.position)
    @input.mark(KEYMAP["up"])
    def move_up(self): self.move_direction(Direction.UP)
    @input.mark(KEYMAP["down"])
    def move_down(self): self.move_direction(Direction.DOWN)
    @input.mark(KEYMAP["newline"])
    def select(self):
        move_cursor(len(self.choices) - self.position, "DOWN")
        return self.position
    @input.mark(KEYMAP["interrupt"])
    def interrupt(self):
        move_cursor(len(self.choices) - self.position, "DOWN")
        raise KeyboardInterrupt
    @input.mark_multiple(*[KEYMAP[str(number)] for number in range(10)])
    def select_row(self):
        index = int(chr(self.current_selection))
        movement = index - self.position
        if index == self.position: return
        if index < len(self.choices):
            if self.position > index: self.move_direction(Direction.UP, -movement)
            elif self.position < index: self.move_direction(Direction.DOWN, movement)
            else: return
        else: return
    def run(self, default_choice: int = 0):
        if self.prompt:
            linebreak()
            forceWrite(self.prompt, "\n")
            if in_colab: forceWrite("Please input a choice index (starting from 0), and press enter", "\n")
            else: forceWrite("Please select a choice using the arrow or number keys, and selecting with enter", "\n")
        self.position = default_choice
        for i in range(len(self.choices)):
            self.print_choice(i)
            forceWrite("\n")
        move_cursor(len(self.choices) - self.position, "UP")
        with cursor.hide():
            while True:
                if in_colab:
                    try: choice = int(builtins.input())
                    except ValueError: choice = default_choice
                else: choice = self.handle_input()
                if choice is not None:
                    reset_cursor()
                    for _ in range(len(self.choices) + 1):
                        move_cursor(1, "UP")
                        clear_line()
                    self.write_choice(choice, "\n")
                    return choice
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
from types import MethodType
import torch.nn as nn
from .imports import is_fp8_available
from .operations import GatheredParameters
def convert_model(model, to_transformer_engine=True, _convert_linear=True, _convert_ln=True):
    if not is_fp8_available(): raise ImportError("Using `convert_model` requires transformer_engine to be installed.")
    import transformer_engine.pytorch as te
    for name, module in model.named_children():
        if isinstance(module, nn.Linear) and to_transformer_engine and _convert_linear:
            has_bias = module.bias is not None
            params_to_gather = [module.weight]
            if has_bias: params_to_gather.append(module.bias)
            with GatheredParameters(params_to_gather, modifier_rank=0):
                if any(p % 16 != 0 for p in module.weight.shape): return
                te_module = te.Linear(module.in_features, module.out_features, bias=has_bias, params_dtype=module.weight.dtype)
                te_module.weight.copy_(module.weight)
                if has_bias: te_module.bias.copy_(module.bias)
                setattr(model, name, te_module)
        elif isinstance(module, nn.LayerNorm) and to_transformer_engine and _convert_ln:
            with GatheredParameters([module.weight, module.bias], modifier_rank=0):
                te_module = te.LayerNorm(module.normalized_shape[0], eps=module.eps, params_dtype=module.weight.dtype)
                te_module.weight.copy_(module.weight)
                te_module.bias.copy_(module.bias)
            setattr(model, name, te_module)
        elif isinstance(module, te.Linear) and not to_transformer_engine and _convert_linear:
            has_bias = module.bias is not None
            new_module = nn.Linear(module.in_features, module.out_features, bias=has_bias, params_dtype=module.weight.dtype)
            new_module.weight.copy_(module.weight)
            if has_bias: new_module.bias.copy_(module.bias)
            setattr(model, name, new_module)
        elif isinstance(module, te.LayerNorm) and not to_transformer_engine and _convert_ln:
            new_module = nn.LayerNorm(module.normalized_shape[0], eps=module.eps, params_dtype=module.weight.dtype)
            new_module.weight.copy_(module.weight)
            new_module.bias.copy_(module.bias)
            setattr(model, name, new_module)
        else: convert_model(module, to_transformer_engine=to_transformer_engine, _convert_linear=_convert_linear, _convert_ln=_convert_ln)
def has_transformer_engine_layers(model):
    if not is_fp8_available(): raise ImportError("Using `has_transformer_engine_layers` requires transformer_engine to be installed.")
    import transformer_engine.pytorch as te
    for m in model.modules():
        if isinstance(m, (te.LayerNorm, te.Linear, te.TransformerLayer)): return True
    return False
def contextual_fp8_autocast(model_forward, fp8_recipe, use_during_eval=False):
    if not is_fp8_available(): raise ImportError("Using `contextual_fp8_autocast` requires transformer_engine to be installed.")
    from transformer_engine.pytorch import fp8_autocast
    def forward(self, *args, **kwargs):
        enabled = use_during_eval or self.training
        with fp8_autocast(enabled=enabled, fp8_recipe=fp8_recipe): return model_forward(*args, **kwargs)
    forward.__wrapped__ = model_forward
    return forward
def apply_fp8_autowrap(model, fp8_recipe_handler):
    if not is_fp8_available(): raise ImportError("Using `apply_fp8_autowrap` requires transformer_engine to be installed.")
    import transformer_engine.common.recipe as te_recipe
    kwargs = fp8_recipe_handler.to_kwargs() if fp8_recipe_handler is not None else {}
    if "fp8_format" in kwargs: kwargs["fp8_format"] = getattr(te_recipe.Format, kwargs["fp8_format"])
    use_during_eval = kwargs.pop("use_autocast_during_eval", False)
    fp8_recipe = te_recipe.DelayedScaling(**kwargs)
    new_forward = contextual_fp8_autocast(model.forward, fp8_recipe, use_during_eval)
    if hasattr(model.forward, "__func__"): model.forward = MethodType(new_forward, model)
    else: model.forward = new_forward
    return model
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import math
from types import MethodType
from typing import Any, Dict, List, Optional, Tuple, Union
from .state import PartialState
from .utils import (calculate_maximum_sizes, convert_bytes, copy_tensor_to_devices, ignorant_find_batch_size, infer_auto_device_map,
is_pippy_available, pad_input_tensors, send_to_device)
def generate_device_map(model, num_processes: int = 1, no_split_module_classes=None, max_memory: dict = None):
    if num_processes == 1: return infer_auto_device_map(model, no_split_module_classes=no_split_module_classes, clean_result=False)
    if max_memory is None:
        model_size, shared = calculate_maximum_sizes(model)
        memory = (model_size + shared[0]) / num_processes
        memory = convert_bytes(memory)
        value, ending = memory.split(" ")
        memory = math.ceil(float(value)) * 1.1
        memory = f"{memory} {ending}"
        max_memory = {i: memory for i in range(num_processes)}
    device_map = infer_auto_device_map(model, max_memory=max_memory, no_split_module_classes=no_split_module_classes, clean_result=False)
    return device_map
def find_pippy_batch_size(args, kwargs):
    found_batch_size = None
    if args is not None:
        for arg in args:
            found_batch_size = ignorant_find_batch_size(arg)
            if found_batch_size is not None: break
    if kwargs is not None and found_batch_size is None:
        for kwarg in kwargs.values():
            found_batch_size = ignorant_find_batch_size(kwarg)
            if found_batch_size is not None: break
    return found_batch_size
def build_pipeline(model, split_points, args, kwargs, num_chunks):
    from torch.distributed.pipelining import ScheduleGPipe, SplitPoint, pipeline
    state = PartialState()
    split_spec = {split_point: SplitPoint.BEGINNING for split_point in split_points}
    pipe = pipeline(model, mb_args=args, mb_kwargs=kwargs, split_spec=split_spec)
    stage = pipe.build_stage(state.local_process_index, device=state.device)
    schedule = ScheduleGPipe(stage, num_chunks)
    return schedule
def pippy_forward(forward, num_chunks, gather_output, *args, **kwargs):
    state = PartialState()
    output = None
    if state.num_processes == 1: output = forward(*args, **kwargs)
    elif state.is_local_main_process:
        found_batch_size = find_pippy_batch_size(args, kwargs)
        if found_batch_size is None: raise ValueError("Could not find batch size from args or kwargs")
        else:
            if found_batch_size != num_chunks:
                args = pad_input_tensors(args, found_batch_size, num_chunks)
                kwargs = pad_input_tensors(kwargs, found_batch_size, num_chunks)
        forward(*args, **kwargs)
    elif state.is_last_process: output = forward()
    else: forward()
    if gather_output: output = copy_tensor_to_devices(output)
    return output
def prepare_pippy(model, split_points: Optional[Union[str, List[str]]] = "auto", no_split_module_classes: Optional[List[str]] = None, example_args: Optional[Tuple[Any]] = (),
example_kwargs: Optional[Dict[str, Any]] = None, num_chunks: Optional[int] = None, gather_output: Optional[bool] = False):
    if not is_pippy_available(): raise ImportError("Using `torch.distributed.pipelining` requires PyTorch 2.4.0 or later.")
    state = PartialState()
    example_args = send_to_device(example_args, "cpu")
    example_kwargs = send_to_device(example_kwargs, "cpu")
    if num_chunks is None: num_chunks = state.num_processes
    if split_points == "auto":
        device_map = generate_device_map(model, num_chunks, no_split_module_classes=no_split_module_classes)
        split_points = []
        for i in range(1, num_chunks): split_points.append(next(k for k, v in device_map.items() if v == i))
    model.hf_split_points = split_points
    stage = build_pipeline(model, split_points, example_args, example_kwargs, num_chunks)
    model._original_forward = model.forward
    model._original_call = model.__call__
    model.pippy_stage = stage
    model.hf_split_points = split_points
    def forward(*args, **kwargs): return pippy_forward(stage.step, num_chunks, gather_output, *args, **kwargs)
    model_forward = MethodType(forward, model)
    forward.__wrapped__ = model_forward
    model.forward = forward
    return model
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
